/* $Header:   F:/source/cwright/cwstart/asm.c_v   1.30   26 Jul 1993 10:16:42   ericj  $ */
/*
 *
 *                Copyright 1991,1992,1993 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN    /* defines the LibMain() */


#define MAX_BUF_LEN    100

static LPSTR   xxx_line_comment = "%";

static LPVOID _xxx_hash_names        = NIL;    /* pointer to the hash table containing */
                                               /* all 'XXX' keyword names. */

static LPVOID _xxx_hash_names_1      = NIL;    /* pointer to optional hash table containing */
static int    _xxx_hash_size_1       = 47;     /* user defined keyword names and size of
                                               /* table. When created these key
                                               /* are set to Color Alternate5 */           
                        

static LPVOID _xxx_hash_names_2      = NIL;    /* pointer to optional hash table containing */
static int    _xxx_hash_size_2       = 47;     /* user defined keyword names and size of
                                               /* table. When created these key
                                               /* are set to Color Alternate6 */              

static LPVOID _xxx_hash_templates    = NIL;    /* pointer to the hash table containing */
                                               /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _xxx_hash_templates.
 */
typedef struct _LanguageTemplate {         
    LPSTR abbrev;
    LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR XXX_INDENT_STR = "\377";

/*
 * _xxx_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _xxx_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _xxx_templates[] = {
//    Abbreviation                       Template
    { "IF"       , "IF    &  THEN" },
    { "WH|WHILE" , "WHILE &" },   
    { NIL        , NIL    },
};

/* Define the list of 'XXX' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _xxx_names[] = {
    "KEYWORD1",
    "KEYWORD2",
    NIL
};


static long _xxx_in_comment( long offset );

void DLL _xxx_init( void );
BOOL DLL _xxx_expand( void );
long DLL _xxx_language_colors_update( long firstline, long lastline );
void DLL _xxx_language_colors( long numlines );
void DLL _xxx_assign_template( LPSTR abbrev, LPSTR template );
void DLL _xxx_add_keyword( LPSTR keyword, int table );
BOOL DLL _xxx_indent( void );
void DLL _xxx_hash_size( int ht, int size );
void DLL _xxx_routines( void);


void DLL 
_init()
{
    _xxx_init();
    
    /* This line causes .yyy files to be treated like .xxx files    */
    ExtAlias(".yyy", ".xxx");
}


/*
 ** XXXSetCommentChar()
 *
 *  Sets what character to use to signal the beginning of a 
 *  "one line" comment.
 */

void DLL
XXXSetCommentChar(LPSTR CharStr)
{
    if (CharStr && *CharStr)
        xxx_line_comment[0] = CharStr[0];
}


/*
 ** _xxx_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_xxx_init( void )
{
    static BOOL firstTime = TRUE;
    int    i;

    if (firstTime)
    {
        /* _xxx language handling functions get called via LibFunctionExec,
         * so they must be exported.
         */
        LibExport( "BOOL _xxx_expand");
        LibExport( "_xxx_assign_template LPSTR LPSTR");
        LibExport( "BOOL _xxx_indent");
        LibExport( "long _xxx_language_colors_update long long");
        LibExport( "_xxx_language_colors long");
        LibExport( "XXXSetCommentChar LPSTR");
        LibExport( "_xxx_add_keyword LPSTR int");
        LibExport( "_xxx_hash_size int int");
		LibExport( "_xxx_routines" );

        firstTime = FALSE;

        /* Create a hash table to hold all of the 'XXX' keyword names.
         */
        _xxx_hash_names = HashCreateTable( 539 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _xxx_names[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _xxx_hash_names, 
                    _xxx_names[i], strlen( _xxx_names[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _xxx_names[i] );
                i++;
            }
        }

        /* Create a hash table to hold all of the 'XXX' template names.
         */
        _xxx_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
        i = 0;

        while ( _xxx_templates[ i ].abbrev != NIL )
        {
            _xxx_assign_template( _xxx_templates[i].abbrev, _xxx_templates[i].template );
            i++;
        }
    }
}


void DLL
_xxx_hash_size( int ht, int size )
{
	if (size > 0)
	{
		if (ht == 1)
			_xxx_hash_size_1 = size;
		else if (ht == 2)
			_xxx_hash_size_2 = size;
	}
}


/*
 ** _xxx_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'XXX' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_xxx_expand( void )
{
    LPVOID htEntry;
    LPSTR  _xxx_string = NULL;
    LPSTR  str;
    int    i = 0;

    MarkSavePos();

    /* Check to see if we are currently positioned inside a 'XXX' comment
     */
    if (_xxx_in_comment( -1L ) != -1)
        goto _xxx_expand_incomplete;

    str = BufReadStr( 65535 );
    while (str[i] && !isspace(str[i]))
        i++;
    str[i] = '\0';

    _xxx_string = StrTrim( str, " \t" );
    StrFree( str );

    if (!_xxx_string)
        goto _xxx_expand_incomplete;

    /* Search for a keyword match in the hash table.
     */
    htEntry = HashFindEntry( _xxx_hash_templates, _xxx_string, strlen( _xxx_string ) );
    StrFree( _xxx_string );

    if (htEntry)
    {
        /* A template exists for the current keyword, delete the keyword
         * and insert the template.
         */
        MarkDropPos( 1 );
        str = *(LPSTR*)htEntry;
        BufDelToEOL();

        _ext_expand_template( str );

    } 
    else
        goto _xxx_expand_incomplete;
    return TRUE;

_xxx_expand_incomplete:
    MarkRestorePos();

    return FALSE;
}

/*
 ** _xxx_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_xxx_language_colors_update( long firstline, long lastline )
{
    long startoff;
    long count;

    if (firstline && lastline)
    {
        /* Begin coloring at the previous comment if
         * currently located in the middle of one.
         */
        _PosInit( 0 );
        if (firstline == 1)
            _PosPrevLine( 0L );
        else
            _PosNextLine( firstline - 1);
        startoff = _PosQOffset();

        /* Clear all attributes for the line range and
         * the reset them within _xxx_language_colors.
         */
        AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
        _PosInit( startoff );
        _xxx_language_colors( lastline - firstline + 1 );
    }

    count = lastline - firstline;
    return(count);
}
                    
void DLL
_xxx_language_colors( long numlines )
{
    long  startoff;      
    long  startline;     
    long  endoff;        
    long  inc;           
    WORD  ch;            
    LPSTR buf;           
    int   i               = 0;              
    int   comment_color   = ColorComments(   -1 );
    int   keyword_color   = ColorKeywords(   -1 );
    int   keyword_color_1 = ColorAlternate5( -1 );
    int   keyword_color_2 = ColorAlternate6( -1 );
    int   quote_color     = ColorAlternate1( -1 );
    int   number_color    = ColorAlternate2( -1 );

    buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

    /* Start at the top of the buffer or current position 
     */
    if (numlines == -1)
    {
        _PosInit( 0 );  
        numlines = 0x7FFFFFFF;
    }
    else
    {
        startoff  = _PosQOffset();
        startline = _PosQLine();
        endoff    = _xxx_in_comment( startoff );
        if (endoff == -1)
            _PosInit( startoff );
        else
        {
            _PosInit( endoff );
            numlines += (startline - _PosQLine());
        }
    }
    ch = _PosCurrentChar();
    startoff = _PosQOffset();


    while ((ch != EOF_CHAR) && (numlines > 0))
    {
        if (ch == '\n')
            numlines--;

        if (ch == '"' || ch == '\'')
        {
            UINT quote = ch;

            startoff = _PosQOffset();

            do 
            {
                ch = _PosNextChar();
                if (ch == '\\')
                {
                    ch = _PosNextChar();
                    if ((ch == quote) || (ch == '\\'))
                        ch = _PosNextChar();
                }
                if ((ch == EOF_CHAR) || (ch == '\n'))
                    break;
            } while (ch != quote);

            if (ch == quote)
                ch = _PosNextChar();

            /* color the string */
            if (quote_color)
            {
                endoff = _PosQOffset();
                _PosInit( startoff );
                _PosSetColor( quote_color, endoff - startoff );
                _PosInit( endoff );
            }
            continue;
        }
        else if (number_color && isdigit( ch ))
        {
            int j = 0;
            UINT lastch = _PosPrevChar();

            if (lastch != EOF_CHAR)
                _PosNextChar();

            startoff = _PosQOffset();
            do
            {
                j++;
                ch = _PosNextChar();
            } while (isxdigit( ch ) || (ch == '.') || (ch == 'x' && j == 1));

            if ((  lastch != ' ' 
                && lastch != '\t' 
                && lastch != '\r' 
                && lastch != '\n' 
                && lastch != EOF_CHAR
                && !ispunct( lastch ))
                || lastch == '_'
                || (ch == '_'))
                continue;

            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( number_color, endoff - startoff );
            _PosInit( endoff );

        }
        else if (isalnum( (BYTE)ch ) || (ch == '_') || (ch == '.'))
        {
            WORD lastch = _PosPrevChar();

            if (lastch != EOF_CHAR)
                _PosNextChar();

            i = 0;
            startoff = _PosQOffset();

            do
            {
                if (i < MAX_BUF_LEN)
                    buf[i++] = (char)ch;
                ch = _PosNextChar();
            } while (isalnum( (BYTE)ch ) || (ch == '_') || (ch == '.'));

            if (strchr( " \t\r\n[(:", lastch ) == NULL)
                i = 0;

            continue;
        }
        else if (i)
        {
            if (isalnum( (BYTE)buf[0] ))
            {
                if (HashFindEntry( _xxx_hash_names, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color, endoff - startoff );
                    _PosInit( endoff );
                }
                else if (HashFindEntry( _xxx_hash_names_1, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color_1, endoff - startoff );
                    _PosInit( endoff );
                }
                else if (HashFindEntry( _xxx_hash_names_2, buf, i ))
                {
                    endoff = _PosQOffset();
                    _PosInit( startoff );
                    _PosSetColor( keyword_color_2, endoff - startoff );
                    _PosInit( endoff );
                }
            }
            i = 0;
        } 

        if (ch == (WORD)*xxx_line_comment)
        {
            startoff = _PosQOffset();

            inc = _PosNextLine( 1 );
            numlines--;
            endoff = _PosQOffset();
            _PosInit( startoff );
            _PosSetColor( comment_color, endoff - startoff - inc );
            _PosInit( endoff );
            ch = _PosCurrentChar();
            continue;
        }
        ch = _PosNextChar();
    }
    MemFree( buf );
}


/*
 ** _xxx_add_keyword()
 *
 *  PARAMETERS:
 *    keyword - the keyword to add to the user-defined keyword table
 *    table   - the 1-based index of the table to which to add the keyword
 *
 *  DESCRIPTION:
 *    Provide a means of adding user-defined keywords for special
 *    coloring
 */
void DLL
_xxx_add_keyword( LPSTR keyword, int table )
{
    if (keyword && *keyword && table >= 0)
    {
        LPSTR XFAR *lpf;
        LPVOID  ht;

        switch( table )
        {
        case 0:
            ht = _xxx_hash_names;
            break;
        case 1:
            if (_xxx_hash_names_1 == NULL)
                _xxx_hash_names_1 = HashCreateTable( _xxx_hash_size_1, sizeof( LPSTR ), 0 );
            ht = _xxx_hash_names_1;
            break;
        case 2:
            if (_xxx_hash_names_2 == NULL)
                _xxx_hash_names_2 = HashCreateTable( _xxx_hash_size_2, sizeof( LPSTR ), 0 );
            ht = _xxx_hash_names_2;
            break;
        }

        lpf = (LPSTR XFAR *)HashGetEntry( ht, keyword, strlen( keyword ) );
        if (lpf)
        {
            *lpf = StrNew( keyword );
        }
    }
}


/*
 ** _xxx_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'XXX' template strings
 *    without changing the code.
 */
void DLL
_xxx_assign_template( LPSTR abbrev, LPSTR template )
{
    _ext_assign_template( _xxx_hash_templates, abbrev, template );
}


/*
 ** _xxx_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'XXX' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in XXX_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_xxx_indent( void )
{
    long   curline;
    long   col;
    LPSTR  _xxx_string;
    LPSTR  str;
    long   matchpos;

    MarkSavePos();

    curline = BufQCurrentLine();

    /* Check to see if we are currently positioned inside a 'XXX' comment
     */
    if (_xxx_in_comment( -1L ) != -1)
        goto _xxx_indent_incomplete;

    /* Not in a comment so indent according to the current 'XXX' construct.
     */
    if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
            || (BufQCurrentLine() != curline) )
        goto _xxx_indent_incomplete;

    /* Read the keyword at the beginning of the current line.
     */
    col = BufQCurrentCol();
    SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
    str = BufReadStr( col - BufQCurrentCol() );
    _xxx_string = StrTrim( str, " \t" );
    StrFree( str );

    /* Set prefix and suffix to \377 to perform a quick lookup of the name
     */
    str = StrNew( "\377" );
    _xxx_string = StrAppend( _xxx_string, str );
    str = StrAppend( str, _xxx_string );
    StrFree( _xxx_string );
    
    matchpos = StrMatch( str, XXX_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
    StrFree( str );

    if (!matchpos)
        /* Not a 'XXX' keyword, do normal indentation. */
        goto _xxx_indent_incomplete;
        
    MarkRestorePos();

    _ext_insert_indented_EOL();
    BufInsertChar( '\t' );
    return TRUE;

    _xxx_indent_incomplete:
        MarkRestorePos();

    return FALSE;
}

/*
 ** _xxx_routines()
 *
 *  DESCRIPTION:
 *    Locates function definitions for selective display mode.
 *    This function gets called by the Codewright API call Routines()
 *    if the current file is a .XXX file.
 *    The code here is taken from _c_routines; modify it as needed.
 */
void DLL
_xxx_routines( void )
{
	int   matches = 0;
	LPSTR str     = "^[a-zA-Z0-9_].*\\((.|\n)*\\)";
	LPSTR stre    = "^[a-zA-Z0-9_].*\\((.|\n)*\\)[ \t\n]*\\c[^ \t\n]";
	long  curline;
	UINT  ch;
	long  lines;

	SysWaitStart();

	MarkSavePos();

	BufSetAscii();
	MovTopBuf();

	while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
	{
		curline = BufQCurrentLine();

		SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

		ch = BufReadChar();
		if (ch != ';')
		{
			if (!matches)
				AttrSetVisible( 0, 0, 0 );
			matches++;
			lines = BufQCurrentLine() - curline
						- !!(ch == '{');
			AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
			MovHome();
		}

		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		BufSetCompact();
		MarkDropPos( 1 );
		MovTopBuf();
	}
	else
		MarkRestorePos();

	SysWaitEnd();

	MsgMessage( matches ? "" : "No routines found." );
}


/*--  Private functions --*/


/*
 ** _xxx_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'XXX' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_xxx_in_comment( long offset )
{
    WORD  ch;

    /* Check to see if we are currently positioned inside a 'XXX' comment
     */
    _PosInit( offset );
    ch = _PosCurrentChar();

    do
    {
        if (ch == '\n')
            return -1;
        else if (ch == (WORD)*xxx_line_comment)
            return _PosQOffset();  /* currrent line is in a comment */
        ch = _PosPrevChar();
    } while (ch != EOF_CHAR);

    return -1L;
}

